import pandas as pd
from os import path

import yaml
from snakemake.utils import validate, Paramspace

validate(config, schema="../schemas/config.schema.yaml")

samples = (
    pd.read_csv(
        config["samples"],
        sep="\t",
        dtype=str,
    )
    .set_index("sample_name", drop=False)
    .sort_index()
)


if "groups" in config:
    groups = pd.read_csv(config["groups"], sep="\t", dtype=str).set_index("group")
    samples = samples.join(groups, on="group")

for name, column in samples.items():
    if column.str.endswith(" ").any() or column.str.startswith(" ").any():
        raise ValueError(
            f"At least one value in column {name} of sample or group sheet has leading "
            "or trailing whitespaces; please ensure that those are removed to avoid "
            "unexpected effects."
        )

def _group_or_sample(row):
    group = row.get("group", None)
    if pd.isnull(group):
        return row["sample_name"]
    return group


samples["group"] = [_group_or_sample(row) for _, row in samples.iterrows()]

validate(samples, schema="../schemas/samples.schema.yaml")

panels = samples["panel"].unique().tolist()


def build_paramspace(paramspace_dicts: list):
    return Paramspace(pd.DataFrame(paramspace_dicts), single_wildcard="paramspace_instance")


paramspace_instances = {
    model: build_paramspace(config["diffexp"]["models"][model].get("paramspace", []))
    for model in config["diffexp"]["models"]
}

def get_paramspace_instance(wildcards):
    return dict(sorted(paramspace_instances[wildcards.model].instance(wildcards).items()))

def get_diffexp_filter_names(model):
    return list(config["diffexp"]["models"][model].get("filter", dict())) + ["nofilter"]

def get_diffexp_query(wildcards):
    if wildcards.diffexp_filter == "nofilter":
        return None

    paramspace_instance = get_paramspace_instance(wildcards)

    diffexp_filter = config["diffexp"]["models"][wildcards.model].get("filter", dict())[wildcards.diffexp_filter]

    queries = []
    for f in diffexp_filter:
        try:
            eval_res = eval(f["if"], dict(), {"paramspace_instance": paramspace_instance})
        except Exception as e:
            raise WorkflowError(f"Error evaluating filter condition {f['if']}: {e}")
        if eval_res:
            queries.append(f["expr"])

    if len(queries) > 1:
        raise WorkflowError(f"More than one differential expression filter query is applicable. Make sure that the provided queries are mutually exclusive.")
    elif len(queries) == 0:
        return None
    else:
        return queries[0]

def get_diffexp_correlation_gene_pairs(model):
    return config["diffexp"]["models"][model].get("correlate_genes", [])

def final_output():
    model_evaluations = [{"model": model, "space": space, "filter": filter} for model, paramspace in paramspace_instances.items() for space in paramspace.instance_patterns for filter in get_diffexp_filter_names(model)]
    final_output = expand(
        [
            "results/datavzrd/diffexp/{item[model]}/{item[space]}/{item[filter]}",
            "results/datavzrd/qc_table/{item[model]}/{item[space]}",
            "results/plots/clustering/{item[model]}/{item[space]}.mds.html",
            "results/plots/qc/{item[model]}/{item[space]}.raw_and_norm.html",
            "results/plots/qc/{item[model]}/{item[space]}.dist_and_filtering.html",
            "results/plots/qc/{item[model]}/{item[space]}.replicates.html",
        ],
        item=model_evaluations,
    )
    final_output.extend(
        expand(
            "results/datavzrd/enrichment/{item[model]}/{item[space]}/{item[filter]}/{enrichr_library}",
            item=model_evaluations,
            enrichr_library=config["enrichment"]["enrichr_libraries"],
        )
    )

    correlations = [
        {"model": item["model"], "space": item["space"], "genes": pair} for item in model_evaluations
        for pair in get_diffexp_correlation_gene_pairs(item["model"])
    ]
    final_output.extend(
        expand(
            "results/plots/gene_correlation/{item[model]}/{item[space]}.{item[genes][gene_a]}_vs_{item[genes][gene_b]}.html",
            item=correlations,
        )
    )
    return final_output


def select_samples(paramspace_instance, samples):
    select = [True] * len(samples)
    for col, value in paramspace_instance.items():
        if not value == "__any__":
            select = select & (samples[col] == value)
    selection = samples.loc[select]
    return selection.sort_values()


def get_other_paramspace_instances(wildcards):
    model = get_model(wildcards)
    paramspace_instance = get_paramspace_instance(wildcards)
    return [
        other_paramspace_instance for other_paramspace_instance in model.get("paramspace", []) 
        if other_paramspace_instance != paramspace_instance
    ]


def get_other_paramspace_instance_descriptors(wildcards):
    paramspace_instance = get_paramspace_instance(wildcards)
    def get_colname(other_paramspace_instance):
        return ", ".join(
            f"{key}_{value}"
            for key, value in other_paramspace_instance.items()
            if paramspace_instance.get(key) != value
        )
    return [get_colname(other_paramspace_instance) for other_paramspace_instance in get_other_paramspace_instances(wildcards)]


def get_other_paramspace_instance_diffexp(wildcards):
    paramspace_instance = get_paramspace_instance(wildcards)
    other_paramspace_instances = get_other_paramspace_instances(wildcards)
    other_paramspace = build_paramspace(other_paramspace_instances)
    return expand(
        "results/differential_expression/{model}/{space}.raw.tsv",
        model=wildcards.model,
        space=other_paramspace.instance_patterns,
    )


def get_differential_logfc_plot(wildcards):
    if get_other_paramspace_instance_descriptors(wildcards):
        return "resources/datavzrd/{model}/{paramspace_instance}/differential_logfc_plot_spec.json"
    else:
        return []


def get_raw_rcc_file_path(wildcards):
    # TODO this is risky because the same filename could in theory occur in multiple different dirs in the sample sheet
    paramspace_instance = get_paramspace_instance(wildcards)
    return select_samples(paramspace_instance, samples).loc[samples["rcc"].str.contains(wildcards.file), "rcc"].squeeze()


def get_paramspace_instance_rcc_files(wildcards):
    paramspace_instance = get_paramspace_instance(wildcards)
    rccs = select_samples(paramspace_instance, samples)["rcc"].apply(lambda x: path.join("resources/rccs/", wildcards.model, wildcards.paramspace_instance, path.basename(x))).tolist()
    if not rccs:
        raise ValueError("No rcc files found. Consider removing this model as there seems to be no data for it.")
    return rccs

def get_report_labels(wildcards, include_model=True):
    labels = {"model": wildcards.model} if include_model else {}
    labels.update(dict(get_paramspace_instance(wildcards)))
    if "diffexp_filter" in wildcards.keys() and wildcards.diffexp_filter != "nofilter":
        labels["filter"] = wildcards.diffexp_filter

    return labels


def get_report_labels_with_n_significant(wildcards, input, include_model=True):
    labels = get_report_labels(wildcards, include_model=include_model)
    table = pd.read_csv(input.table, sep="\t")
    fdr_cols = table.columns.str.startswith("FDR")
    labels["# significant"] = (table.loc[:, fdr_cols] <= 0.05).any(axis="columns").sum()
    return labels


def get_correlation_report_labels(wildcards):
    labels = get_report_labels(wildcards, include_model=False)
    labels["gene_a"] = wildcards.gene_a
    labels["gene_b"] = wildcards.gene_b
    return labels


def get_model(wildcards):
    return config["diffexp"]["models"][wildcards.model]

wildcard_constraints:
    model="|".join(config["diffexp"]["models"]),
    panel="|".join(panels),