log <- file(snakemake@log[[1]], open="wt")
sink(log)
sink(log, type="message")

library("sjmisc")
library("edgeR")
library("tidyverse")

data <- read_tsv(
        snakemake@input[["counts"]]
    ) %>% 
    drop_na(!c(name, type, accession)) # drop NA rows (removed due to background correction)

normcounts <- data %>%
    select(!c(name, type)) %>%
    column_to_rownames(var = "accession")

features <- data %>%
    select(c(name, type, accession)) %>%
    column_to_rownames(var = "accession")

model <- snakemake@params[["model"]]

samples <- read_csv(
        snakemake@input[["samples"]]
    ) %>%
    # make everything except the index, sample name and path string a factor
    mutate_at(vars(c(-sample_name, -all_of(model[["continuous_cols"]]))), list(~factor(.)))

# retrieve the model formula
formula <- as.formula(model[["formula"]])

# extract variables from the formula and unnest any nested variables
variables <- labels(terms(formula)) %>%
                strsplit('[:*]') %>%
                unlist()

expr_variables <- variables %>% str_subset("^expr_") %>% str_remove("^expr_")
sample_sheet_variables <- variables %>% str_subset("^(?!expr_)")

# remove samples with an NA value in any of the columns
# relevant for edgeR under the current model
samples <- samples %>%
        drop_na(c(sample_name, all_of(sample_sheet_variables)))
primary_variable <- model[["primary_variable"]]
base_level <- model[["base_level"]]
# TODO migrate this to tidyverse
# Ensure that primary variable factors are sorted such that base_level comes first.
# This is important for fold changes, effect sizes to have the expected sign.
samples[, primary_variable] <- relevel(samples[, primary_variable, drop=TRUE], base_level)
samples <- samples %>% select(c(sample_name, all_of(sample_sheet_variables)))

# add expression variables to samples
if (length(expr_variables) > 0) {
    expr_variables <- rotate_df(data %>% filter(name %in% expr_variables) %>% select(c(-accession, -type)))
    colnames(expr_variables) <- expr_variables[1, ]
    expr_variables <- expr_variables[-1, , drop=FALSE]
    # prepent gene columns with expr_ again
    colnames(expr_variables) <- paste0("expr_", colnames(expr_variables))
    expr_variables <- expr_variables %>% 
        rownames_to_column(var = "sample_name") %>% 
        as_tibble() %>%
        mutate(across(-sample_name, as.double))

    samples <- samples %>%
        left_join(
            expr_variables,
            by = "sample_name"
        )
}

# Remove variables that have only one value
n_values <- summarize_all(samples, n_distinct)
if (TRUE %in% (n_values <= 1)) {
    stop(paste0("Unable to generate design matrix because formula refers to variables with only one value: ", paste(colnames(n_values)[n_values <= 1], collapse = ", ")))
}

# restrict normcounts to selected samples
normcounts <- normcounts %>% select(all_of(pull(samples, sample_name)))

dge_list <- DGEList(counts=normcounts, genes=features)

design <- model.matrix(
    formula,
    data=samples
)

dge_list <- estimateDisp(dge_list, design)
fit <- glmQLFit(dge_list, design)
qlf <- glmQLFTest(fit, coef = model[["coef"]])

saveRDS(dge_list, file = snakemake@output[["dgelist"]])

table <- rownames_to_column(
        topTags(qlf, n = 100000)$table,
        var = "accession"
    ) %>%
    mutate_if(is.numeric, signif, 3) %>%
    rename(pvalue=PValue)

write_tsv(
    table,
    snakemake@output[["diffexp_table"]])

pdf(file = snakemake@output[["bcv"]])
plotBCV(dge_list)
dev.off()

pdf(file = snakemake@output[["md"]])
plotMD(qlf)
abline(h = c(-1,1), col = "grey")
dev.off()

pdf(file = snakemake@output[["disp"]])
plotQLDisp(fit)
dev.off()