import sys

sys.stderr = open(snakemake.log[0], "w")

import pandas as pd
import altair as alt
from altair import datum


feature_cols = ["name", "accession", "type"]
counts = pd.read_csv(snakemake.input["counts"], sep="\t", index_col=feature_cols)
annotation = pd.read_csv(snakemake.input["ann"], sep=",", index_col="sample_name")

# remove replicates and only keep the first sample
# Note: when a specific replicate shall be removed, this should be done via the sample sheet
replicates = annotation.loc[annotation[["sample_type", "group"]].duplicated()].index.values
replicate_to_remove = counts.columns.isin(replicates)
removed_replicates = counts.columns[replicate_to_remove].values
filtered_counts = counts.loc[:, ~replicate_to_remove]

# remove samples with low mean expression
expression_to_keep = filtered_counts.apply(
    lambda col: (~(col == 0.0)).sum() / filtered_counts.shape[0] >= snakemake.params.min_fraction_nonzero
).values
removed_low_expression = filtered_counts.columns[~expression_to_keep].values
filtered_counts = filtered_counts.loc[:, expression_to_keep]

# filter annotation accordingly
annotation = annotation.loc[filtered_counts.columns]

filtered_counts.to_csv(snakemake.output["counts"], sep="\t")
annotation.index.name = "sample_name"
annotation.to_csv(snakemake.output["ann"], sep=",")

# plot QC
stacked_counts = pd.DataFrame(counts.stack())
stacked_counts.index.rename("sample_name", -1, inplace=True)
stacked_counts.columns = ["expression"]
stacked_counts["filter-status"] = "kept"
stacked_counts.loc[(slice(None), slice(None), slice(None), removed_replicates), "filter-status"] = "removed replicates"
stacked_counts.loc[(slice(None), slice(None), slice(None), removed_low_expression), "filter-status"] = "removed low expression samples"

chart = alt.Chart(stacked_counts.reset_index()).transform_filter(datum.expression <= 1000.0).transform_window(
    cumulative_count="count()",
    sort=[{"field": "expression", "order": "ascending"}],
    groupby=["sample_name"],
).mark_line(tooltip=alt.TooltipContent(content="data")).encode(
    alt.X("expression:Q"),
    alt.Y("cumulative_count:Q", title="CDF", axis=alt.Axis(labels=False, values=[0], grid=False, ticks=False)),
    alt.Color("sample_name:N", scale=alt.Color(range=["#1f77b4"]), legend=None),
    tooltip=["sample_name:N"],
    opacity=alt.value(0.5),
).facet("filter-status:N")

chart.save(snakemake.output["plt"])

