log <- file(snakemake@log[[1]], open="wt")
sink(log)
sink(log, type="message")

library("NanoTube")
library("tidyverse")
library("ruv")

# best to keep this fixed to nSolver:
# * nSolver gives plausible normalization results, including with background removal
# * RUVg seems to be incompatible with background removal (results in very low expression values and some negative values)
# * RUVIII doesn't give raw counts in the 'list' output.format mode
normalization <- "nSolver"
id_col <- "rcc"
bg_type <- "threshold"
bg_sds <- 2
bg_proportion <- 0.5
bg_p_val <- 0.01
bg_subtract <- TRUE

group_col <- snakemake@params[["model"]][["primary_variable"]]

if (normalization == "RUVg" && bg_subtract == TRUE) {
  cli_abort("Do not use RUVg normalization with bgSubtract = TRUE, as resulting values are somehow strongly deflated.")
}

exprs_list <- processNanostringData(
  nsFile = snakemake@params[["input_folder"]],
  sampleTab = snakemake@input[["pheno_data"]],
  includeQC = TRUE,
  output.format = "list",
  idCol = id_col,
  groupCol = group_col,
  normalization = normalization,
  bgType = bg_type,
  bgThreshold = bg_sds,
  bgProportion = bg_proportion,
  bgPVal = bg_p_val,
  bgSubtract = bg_subtract
)

# generate RLE (Relative Log Expression) plot, see:
# https://www.bioconductor.org/packages/release/bioc/vignettes/NanoTube/inst/doc/NanoTube.html#rle-plots
pdf(
  snakemake@output[["rle_plot"]],
  width=10,
  height=5
)
ruv_rle(
  t(log2(exprs_list$exprs)) %>%
    # replace actual 0 expression values with veeery low expression values
    # in log space, to avoid -Inf values that fail ruv_rle
    ifelse(. == -Inf, -1000000000, .),
  ylim = c(-2, 2)
)
dev.off()

positive_qc <- positiveQC(exprs_list)$tab %>%
  as_tibble(rownames="sample") %>%
  select(-Sample) %>%
  rename(
    "positive control scaling factor"=Scale.Factor#,
#    "R squared"=R.squared
    )
negative_qc <- negativeQC(exprs_list, interactive.plot = FALSE)$tab %>%
  as_tibble(rownames="sample") %>%
  rename(
    "mean of negative controls" = "Mean (Neg)",
    "standard deviation of negative controls" = "sd (Neg)",
    "maximum among negative controls" = "Max (Neg)"
  ) %>%
  separate_wider_regex(
    `Genes below BG (%)`,
    c(
      `number genes below background` = "\\d+",
      "\\s\\(",
      `percentage genes below background` = "[\\d\\.]+",
      "%\\)"
    )
  )
housekeeping_qc <- exprs_list$hk.scalefactors %>%
  enframe(
    name="sample",
    value="housekeeping scaling factor"
  )

qc_table <- positive_qc %>%
  full_join(negative_qc) %>%
  full_join(housekeeping_qc) %>%
  mutate(
    sample = str_remove(sample, "\\.RCC$")
  )
write_tsv(
  x = qc_table,
  file = snakemake@output[["qc_table"]]
)

format_exprs <- function(source_type) {
  exprs <- exprs_list[[source_type]]
  # raise to the power of 2, because the RUVIII normalization converts to log2
  # but we want just counts downstream
  if(source_type == "exprs" && exprs_list$normalization == "RUVIII") {
    exprs <- 2 ^ exprs
  }

  colnames(exprs) <- exprs_list$samples[colnames(exprs_list[[source_type]]), "sample_name"]

  exprs %>%
    as_tibble(rownames = "Accession") %>%
    full_join(exprs_list$dict.raw %>% as_tibble()) %>%
    rename(
      accession = Accession,
      name = Name,
      type = CodeClass
    ) %>%
    relocate(name, accession, type)
}

raw_counts <- format_exprs("exprs.raw")
norm_counts <- format_exprs("exprs")

# RUVg normalization only happens after backgroung removal and can
# in combination with background removal lead to slightly negative
# normalized counts, which are not allowed for differential
# expression analysis
if ( exprs_list$normalization == "RUVg" ) {
  norm_counts <- norm_counts %>%
    mutate(
      across(
        !c(accession, type, name),
        ~ if_else(. < 0, 0, . )
      )
    )
}

norm_counts %>% write_tsv(snakemake@output[["norm_counts_matrix"]])

all_counts <- raw_counts %>%
  add_column(count_type = "raw") %>%
  bind_rows(
    norm_counts %>%
      add_column(count_type = "normalized")
  ) %>%
  pivot_longer(
    !c(type, name, accession, count_type),
    names_to = "sample",
    values_to = "expression"
  ) %>%
  write_tsv(snakemake@output[["exprs_and_norm"]])