import sys

sys.stderr = open(snakemake.log[0], "w")

import altair as alt
import pandas as pd

# load counts
feature_cols = ["name", "accession", "type"]
counts = pd.DataFrame(pd.read_csv(snakemake.input["counts"], sep="\t", index_col=feature_cols).stack())
counts.index.rename("sample_name", -1, inplace=True)
counts.columns = ["expression"]

# load annotation
annotation = pd.read_csv(snakemake.input["ann"], sep=",", index_col="sample_name")

# check that there are no samples with more than one replicate
has_more_than_one_replicate = (annotation[["sample_type", "group"]].duplicated().groupby("sample_name").sum() > 1).any()
if has_more_than_one_replicate:
    raise ValueError("More than one replicate per sample found. This is not supported.")

# add annotation to counts
ann_cols = ["sample_type", "group"]
ann_values = annotation.loc[counts.index.get_level_values("sample_name"), ann_cols].values
counts[ann_cols] = ann_values
counts = counts.reset_index().set_index(feature_cols + ["sample_name"] + ann_cols)

def select_by_sample_type(sample_type):
    """Select counts for a given sample type and return a dataframe with primary and secondary samples."""
    sample_type_ann = annotation.loc[annotation["sample_type"] == sample_type]

    # restrict to groups with replicates
    groups_with_replicates = sample_type_ann.loc[sample_type_ann[["sample_type", "group"]].duplicated(), "group"].values
    group_counts = counts.loc[:, :, :, :, :, groups_with_replicates]
    sample_type_ann = sample_type_ann.loc[sample_type_ann["group"].isin(groups_with_replicates)]

    replicates = sample_type_ann[["sample_type", "group"]].duplicated()
    primary_samples = sample_type_ann.index[~replicates].values
    secondary_samples = sample_type_ann.index[replicates].values

    def select_counts(samples, name):
        selection = group_counts.loc[:, :, :, samples].reset_index(level=["sample_name"])
        selection.columns = [f"{name}_sample", f"{name}_expression"]
        return selection

    return pd.concat([select_counts(primary_samples, "primary"), select_counts(secondary_samples, "secondary")], axis="columns")

# select counts for each sample type and concatenate
counts = pd.concat([select_by_sample_type(sample_type) for sample_type in annotation["sample_type"].unique()])

# reset multiindex such that altair is happy
counts = counts.reset_index()

# plot
chart = alt.Chart(counts).mark_point(tooltip=alt.TooltipContent(content="data")).encode(
    alt.X("primary_expression:Q", scale=alt.Scale(type="symlog")),
    alt.Y("secondary_expression:Q", scale=alt.Scale(type="symlog")),
    color="sample_type:N",
).facet("group:N", columns=4)

chart.save(snakemake.output[0])